package org.simantics.document.server.request;

import java.util.Collections;
import java.util.Map;

import org.simantics.databoard.Bindings;
import org.simantics.db.ReadGraph;
import org.simantics.db.Resource;
import org.simantics.db.Statement;
import org.simantics.db.common.procedure.adapter.TransientCacheListener;
import org.simantics.db.common.request.IndexRoot;
import org.simantics.db.common.request.UnaryRead;
import org.simantics.db.common.utils.NameUtils;
import org.simantics.db.exception.DatabaseException;
import org.simantics.db.layer0.scl.AbstractExpressionCompilationContext;
import org.simantics.db.layer0.util.RuntimeEnvironmentRequest2;
import org.simantics.db.layer0.variable.Variable;
import org.simantics.layer0.Layer0;
import org.simantics.scl.compiler.common.names.Name;
import org.simantics.scl.compiler.constants.StringConstant;
import org.simantics.scl.compiler.elaboration.expressions.EApply;
import org.simantics.scl.compiler.elaboration.expressions.EConstant;
import org.simantics.scl.compiler.elaboration.expressions.ELiteral;
import org.simantics.scl.compiler.elaboration.expressions.EVariable;
import org.simantics.scl.compiler.elaboration.expressions.Expression;
import org.simantics.scl.compiler.elaboration.modules.SCLValue;
import org.simantics.scl.compiler.environment.Environment;
import org.simantics.scl.compiler.environment.Environments;
import org.simantics.scl.compiler.environment.specification.EnvironmentSpecification;
import org.simantics.scl.compiler.runtime.RuntimeEnvironment;
import org.simantics.scl.compiler.top.SCLExpressionCompilationException;
import org.simantics.scl.compiler.types.Type;
import org.simantics.scl.compiler.types.Types;
import org.simantics.scl.compiler.types.kinds.Kinds;
import org.simantics.scl.runtime.SCLContext;
import org.simantics.scl.runtime.function.Function1;
import org.simantics.structural2.scl.ComponentTypeProperty;
import org.simantics.structural2.scl.FindPossibleComponentTypeRequest;
import org.simantics.structural2.scl.ReadComponentTypeInterfaceRequest;
import org.simantics.utils.datastructures.Pair;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

public class ServerSCLValueRequest extends ServerSCLValueRequestBase<org.simantics.document.server.request.ServerSCLValueRequest.CompilationContext> {

	private final Pair<Resource,Resource> componentTypeAndRoot;
	private final Resource literal;
    protected String possibleExpectedValueType;

	public static class CompilationContext extends AbstractExpressionCompilationContext {
		public final Map<String, ComponentTypeProperty> propertyMap;

		public CompilationContext(RuntimeEnvironment runtimeEnvironment,
				Map<String, ComponentTypeProperty> propertyMap) {
			super(runtimeEnvironment);
			this.propertyMap = propertyMap;
		}
	}

	private ServerSCLValueRequest(Pair<Resource,Resource> componentTypeAndRoot, Resource literal, String possibleExpectedValueType) {
		assert(literal != null);
		this.literal = literal;
		this.componentTypeAndRoot = componentTypeAndRoot;
		this.possibleExpectedValueType = possibleExpectedValueType;
	}

	public ServerSCLValueRequest(ReadGraph graph, Variable context) throws DatabaseException {
		this(getComponentTypeAndRoot(graph, context), context.getRepresents(graph), resolveExpectedValueType(graph, context.getPredicateResource(graph)));
	}

	public ServerSCLValueRequest(ReadGraph graph, Resource s, Resource o, Resource p) throws DatabaseException {
	    this(getComponentTypeAndRoot(graph, s, o), o, resolveExpectedValueType(graph, p));
	}

	private static Pair<Resource,Resource> getComponentTypeAndRoot(ReadGraph graph, Variable property)  throws DatabaseException {
		Variable parent = property.getParent(graph);
		Resource represents = parent.getRepresents(graph);
		if(represents != null) {
			Resource type = graph.syncRequest(new FindPossibleComponentTypeRequest(represents));
			if(type != null) {
				Resource root = graph.syncRequest(new IndexRoot(type));
				return Pair.make(type, root);
			}
		}
		parent = parent.getParent(graph);
		Resource root = graph.syncRequest(new IndexRoot(property.getRepresents(graph)));
		return Pair.make(parent.getType(graph), root);
	}

	public static Object compileAndEvaluate(ReadGraph graph, Variable context) throws DatabaseException {
		SCLContext sclContext = SCLContext.getCurrent();
		Object oldGraph = sclContext.get("graph");
		try {
			Function1<Object,Object> exp = compile(graph, context);
			sclContext.put("graph", graph);
			return exp.apply(context);
		} catch (DatabaseException e) {
			throw (DatabaseException)e;
		} catch (Throwable t) {
			throw new DatabaseException(t);
		} finally {
			sclContext.put("graph", oldGraph);
		}
	}

	public static Function1<Object, Object> compile(ReadGraph graph, Variable context) throws DatabaseException {
		return graph.syncRequest(new ServerSCLValueRequest(graph, context), TransientCacheListener.instance());
	}

	public static Function1<Object, Object> compile(ReadGraph graph, Resource s, Resource o, Resource p) throws DatabaseException {
		return graph.syncRequest(new ServerSCLValueRequest(graph, s, o, p), TransientCacheListener.instance());
	}

	@Override
	protected String getExpressionText(ReadGraph graph)
			throws DatabaseException {
		Layer0 L0 = Layer0.getInstance(graph);
		return graph.getRelatedValue(literal, L0.SCLValue_expression, Bindings.STRING);
	}

	protected RuntimeEnvironmentRequest2 getRuntimeEnvironmentRequest(Resource componentType, Resource indexRoot) {
		return new RuntimeEnvironmentRequest2(componentType, indexRoot) {
			@Override
			protected void fillEnvironmentSpecification(
					EnvironmentSpecification environmentSpecification) {
			}
		};
	}

	@Override
	protected CompilationContext getCompilationContext(ReadGraph graph)
			throws DatabaseException {
		return graph.syncRequest(new UnaryRead<Pair<Resource,Resource>,CompilationContext>(componentTypeAndRoot) {
			@Override
			public CompilationContext perform(ReadGraph graph)
					throws DatabaseException {
				RuntimeEnvironment runtimeEnvironment = graph.syncRequest(getRuntimeEnvironmentRequest(parameter.first, parameter.second));
				Map<String, ComponentTypeProperty> propertyMap;
				if (parameter.first != null) {
				    propertyMap =
						graph.syncRequest(new ReadComponentTypeInterfaceRequest(parameter.first, runtimeEnvironment.getEnvironment()),
								TransientCacheListener.<Map<String, ComponentTypeProperty>>instance());
				} else {
				    // TODO: Antti to consider
				    // To handle procedural user components
				    propertyMap = Collections.emptyMap();
				}
				return new CompilationContext(runtimeEnvironment, propertyMap);
			}
		});
	}

	@Override
	protected Type getContextVariableType() {
		return VARIABLE; 
	}

	private static Expression accessInputVariable(Environment environment,
			org.simantics.scl.compiler.elaboration.expressions.Variable contextVariable) {
		SCLValue variableParentFunction = environment.getValue(VARIABLE_PARENT);
		return new EApply(new EConstant(variableParentFunction),
				new EApply(new EConstant(variableParentFunction),
						new EVariable(contextVariable)));
	}

    protected static Name PROPERTY_VALUE_CACHED = Name.create("Document/All", "propertyValueCached");

    protected static Expression getProperty(Environment environment, Expression variable, String propertyName, Type type) {
        return new EApply(
                new EConstant(environment.getValue(FROM_DYNAMIC), type),
                new EApply(
                        new EConstant(environment.getValue(PROPERTY_VALUE_CACHED), Types.DYNAMIC),
                        variable,
                        new ELiteral(new StringConstant(propertyName))));
    }
    
    protected static Expression getPropertyFlexible(Environment environment, Expression variable, String propertyName, Type type) {
        return makeTypeFlexible(environment, getProperty(environment, variable, propertyName, type), type);
    }
	
	protected static Expression standardGetProperty(
			Environment environment,
			org.simantics.scl.compiler.elaboration.expressions.Variable contextVariable,
			String name,
			Type type) {
		return getPropertyFlexible(environment, accessInputVariable(environment, contextVariable), name, type);
	}


	@Override
	protected Expression getVariableAccessExpression(
			ReadGraph graph,
			CompilationContext context,
			org.simantics.scl.compiler.elaboration.expressions.Variable contextVariable,
			String name) throws DatabaseException {
		ComponentTypeProperty property = context.propertyMap.get(name);
		if(property != null)
			return standardGetProperty(
					context.runtimeEnvironment.getEnvironment(),
					contextVariable,
					name,
					property.type == null ? Types.metaVar(Kinds.STAR) : property.type);
		else
			return getSpecialVariableAccessExpression(graph, context, contextVariable, name);
	}

	protected Expression getSpecialVariableAccessExpression(ReadGraph graph,
			CompilationContext context,
			org.simantics.scl.compiler.elaboration.expressions.Variable contextVariable,
			String name) throws DatabaseException {
		if(name.equals("input")) {
			Environment environment = context.runtimeEnvironment.getEnvironment();
			return accessInputVariable(environment, contextVariable);
		}
		else if(name.equals("self"))
			return new EVariable(contextVariable);
		else
			return null;
	}
	
    @Override
    protected Type getExpectedType(ReadGraph graph, CompilationContext context) throws DatabaseException {
        return super.getExpectedType(graph, context);
    }


	@Override
	public int hashCode() {
		return 31*(31*getClass().hashCode() + literal.hashCode()) + componentTypeAndRoot.hashCode();
	}

	@Override
	public boolean equals(Object obj) {
		if(this == obj)
			return true;
		if(obj == null || obj.getClass() != getClass())
			return false;
		ServerSCLValueRequest other = (ServerSCLValueRequest)obj;
		return literal.equals(other.literal) && componentTypeAndRoot.equals(other.componentTypeAndRoot);
	}

    public static Function1<Object, Object> validate(ReadGraph graph, Variable context) throws DatabaseException {
        return graph.syncRequest(new ServerSCLValueValidationRequest(graph, context), TransientCacheListener.instance());
    }

    @Override
    protected String getContextDescription(ReadGraph graph) throws DatabaseException {
        Layer0 L0 = Layer0.getInstance(graph);
        Statement possibleOwner = graph.getPossibleStatement(literal, L0.IsOwnedBy);
        if(possibleOwner != null) {
            String uri = graph.getPossibleURI(possibleOwner.getObject());
            if(uri != null) {
                String propertyName = NameUtils.getSafeName(graph, graph.getInverse(possibleOwner.getPredicate()));
                return uri + "#" + propertyName;
            }
        }
        return super.getContextDescription(graph);
    }
    
    public static class ServerSCLValueValidationRequest extends ServerSCLValueRequest {

        private static final Logger LOGGER = LoggerFactory.getLogger(ServerSCLHandlerValueRequest.class);

        public ServerSCLValueValidationRequest(ReadGraph graph, Variable context) throws DatabaseException {
            super(graph, context);
        }

        @Override
        protected Type getExpectedType(ReadGraph graph, CompilationContext context) throws DatabaseException {
            if(possibleExpectedValueType != null) {
                try {
                    return Environments.getType(context.runtimeEnvironment.getEnvironment(), possibleExpectedValueType);
                } catch (SCLExpressionCompilationException e) {
                    LOGGER.error("Could not get type for " + String.valueOf(possibleExpectedValueType), e);
                }
            }
            return super.getExpectedType(graph, context);
        }
    }
}
